"""Basic tests of cosmotile algorithm."""
from __future__ import annotations

from typing import Any
from typing import Sequence

import numpy as np
import pytest
from astropy import units as un
from astropy.cosmology import Planck18
from scipy.spatial.transform import Rotation as R  # noqa: N817

import cosmotile as cmt
from cosmotile import make_healpix_lightcone_slice
from cosmotile import make_lightcone_slice


def test_make_lightcone_slice_inputs() -> None:
    """Simple tests that bad inputs raise appropriate errors."""
    coeval = np.random.uniform(size=(10, 10, 10))
    lat = np.random.uniform(size=11) * np.pi - np.pi / 2
    lon = np.random.uniform(size=11) * 2 * np.pi

    def call(
        coeval: np.ndarray | Sequence[np.ndarray] = coeval,
        latitude: np.ndarray = lat,
        longitude: np.ndarray = lon,
        distance_to_shell: float = 1.0,
        **kw: Any,
    ) -> None:
        next(
            make_lightcone_slice(
                coevals=[coeval] if isinstance(coeval, np.ndarray) else coeval,
                latitude=latitude,
                longitude=longitude,
                distance_to_shell=distance_to_shell,
                **kw,
            )
        )

    with pytest.raises(ValueError, match="all coevals must have three dimensions"):
        call(coeval=coeval[0])

    with pytest.raises(
        ValueError, match="latitude and longitude must have the same shape"
    ):
        call(latitude=np.concatenate((lat, [0])))  # type: ignore[arg-type]

    with pytest.raises(ValueError, match="latitude and longitude must be 1D arrays"):
        call(latitude=np.zeros((11, 11)), longitude=np.zeros((11, 11)))

    with pytest.raises(ValueError, match="latitude must be between -pi/2 and pi/2"):
        call(latitude=np.arange(11))

    with pytest.raises(ValueError, match="longitude must be between 0 and 2pi"):
        call(longitude=-1 * np.ones(11))

    with pytest.raises(ValueError, match="distance_to_shell must be positive"):
        call(distance_to_shell=-1)

    with pytest.raises(
        ValueError, match="interpolation_order must be in the range 0-5"
    ):
        call(interpolation_order=1000)

    with pytest.raises(TypeError, match="interpolation_order must be an integer"):
        call(interpolation_order=1.0)

    with pytest.raises(ValueError, match="all coevals must have the same shape"):
        call(coeval=[coeval[:8, :8, :8], coeval])

    with pytest.raises(ValueError, match="origin must be a sequence of length 3"):
        call(origin=(1, 2, 3, 4))


@pytest.mark.parametrize("distance_to_shell", [5, 20, np.pi])
@pytest.mark.parametrize("origin", [(0, 0, 0), (3, 6, -1), (1000, -1000, np.pi)])
@pytest.mark.parametrize(
    "rotation",
    [
        R.from_quat([0, 0, np.sin(np.pi / 4), np.cos(np.pi / 4)]),
        None,
    ],
)
def test_uniform_box(
    distance_to_shell: float,
    origin: tuple[float, float, float],
    rotation: R | None,
) -> None:
    """Test that a uniform box gives uniform interpolated values."""
    coeval = np.ones((10, 10, 10))
    lat = np.random.uniform(size=11) * np.pi - np.pi / 2
    lon = np.random.uniform(size=11) * 2 * np.pi

    shell = list(
        make_lightcone_slice(
            coevals=coeval,
            distance_to_shell=distance_to_shell,
            origin=origin,
            rotation=rotation,
            latitude=lat,
            longitude=lon,
        )
    )[0]
    assert np.allclose(shell, 1, rtol=1e-8)


@pytest.mark.parametrize("distance_to_shell", [5, 20, np.pi])
@pytest.mark.parametrize("origin", [(0, 0, 0), (3, 6, -1), (1000, -1000, np.pi)])
@pytest.mark.parametrize(
    "rotation",
    [
        R.from_quat([0, 0, np.sin(np.pi / 4), np.cos(np.pi / 4)]),
        None,
    ],
)
def test_random_uniform_box(
    distance_to_shell: float,
    origin: tuple[float, float, float],
    rotation: R | None,
) -> None:
    """Test that a random uniform box doesn't yield answers bigger than the maximum."""
    coeval = np.random.uniform(size=(12, 13, 14))
    lat = np.random.uniform(size=11) * np.pi - np.pi / 2
    lon = np.random.uniform(size=11) * 2 * np.pi

    shell = next(
        make_lightcone_slice(
            coevals=coeval,
            distance_to_shell=distance_to_shell,
            origin=origin,
            rotation=rotation,
            latitude=lat,
            longitude=lon,
        )
    )
    assert np.all(shell <= 1)


@pytest.mark.parametrize("distance_to_shell", [5, 20, np.pi])
@pytest.mark.parametrize("origin", [(0, 0, 0), (3, 6, -1), (1000, -1000, np.pi)])
@pytest.mark.parametrize(
    "rotation",
    [
        R.from_quat([0, 0, np.sin(np.pi / 4), np.cos(np.pi / 4)]),
        None,
    ],
)
def test_healpix(
    distance_to_shell: float,
    origin: tuple[float, float, float],
    rotation: R | None,
) -> None:
    """Test that a random uniform box doesn't yield answers bigger than the maximum."""
    coeval = np.random.uniform(size=(12, 13, 14))

    shell = next(
        make_healpix_lightcone_slice(
            coevals=coeval,
            distance_to_shell=distance_to_shell,
            origin=origin,
            rotation=rotation,
            nside=16,
        )
    )
    assert np.all(shell <= 1)


def test_distance_to_shell() -> None:
    """Test the get_distance_to_shell_from_redshift function."""
    assert cmt.get_distance_to_shell_from_redshift(
        z=1, cell_size=1 * un.Mpc, cosmo=Planck18
    ) < cmt.get_distance_to_shell_from_redshift(
        z=2, cell_size=1 * un.Mpc, cosmo=Planck18
    )

    assert cmt.get_distance_to_shell_from_redshift(
        z=1, cell_size=1 * un.Mpc, cosmo=Planck18
    ) == 2 * cmt.get_distance_to_shell_from_redshift(
        z=1, cell_size=2 * un.Mpc, cosmo=Planck18
    )


def test_lightcone_slice_vector_field() -> None:
    """Test the make_lightcone_slice_vector_field function."""
    coeval = np.ones((10, 10, 10))
    distance_to_shell = 1.0 * un.pixel
    lat = np.zeros(21)
    lon = np.linspace(0, 2 * np.pi, 21)

    # put points at the poles for testing
    lat = np.append(lat, [np.pi / 2, -np.pi / 2])
    lon = np.append(lon, [0, 0])

    # all displacement is in x-direction
    rsds = [
        [
            np.ones_like(coeval) * un.pixel,
            np.zeros_like(coeval) * un.pixel,
            np.zeros_like(coeval) * un.pixel,
        ]
    ]

    interpolator = cmt.make_lightcone_slice_interpolator(
        latitude=lat, longitude=lon, distance_to_shell=distance_to_shell
    )

    with pytest.raises(
        ValueError,
        match="coeval_vector_fields must be a sequence of 3-tuples. Got length",
    ):
        next(
            cmt.make_lightcone_slice_vector_field(
                interpolator=interpolator,
                coeval_vector_fields=[rsds[0][:2]],
            )
        )

    with pytest.raises(
        ValueError, match="all coeval vector fields must have the same shape."
    ):
        next(
            cmt.make_lightcone_slice_vector_field(
                interpolator=interpolator,
                coeval_vector_fields=[[rsds[0][0][1:], rsds[0][1], rsds[0][2]]],
            )
        )

    los = list(
        cmt.make_lightcone_slice_vector_field(
            interpolator=interpolator,
            coeval_vector_fields=rsds,
        )
    )[0]

    assert los[0] == -1 * un.pixel
    assert los[10] == 1 * un.pixel
    assert los[20] == -1 * un.pixel
    assert np.isclose(los[21], 0 * un.pixel)
    assert np.isclose(los[22], 0 * un.pixel)


def test_apply_rsds() -> None:
    """Test the apply_rsds function."""
    field = np.ones((2, 10))
    losv = np.ones((2, 10)) * un.pixel
    losv[1] = -1 * un.pixel
    distance = np.array([10, 11]) * un.pixel

    with pytest.raises(ValueError, match="field must have at least 2 slices"):
        cmt.apply_rsds(
            field=field[:1],
            los_displacement=losv[:1],
            distance=distance[:1],
            n_subcells=1,
        )

    with pytest.raises(
        ValueError, match="field and los_displacement must have the same shape"
    ):
        cmt.apply_rsds(
            field=field[:, :9],
            los_displacement=losv,
            distance=distance,
            n_subcells=1,
        )

    with pytest.raises(
        ValueError, match="field and distance must have the same number"
    ):
        cmt.apply_rsds(
            field=field,
            los_displacement=losv,
            distance=distance[:1],
            n_subcells=1,
        )

    # Here, the close slice all moves inward, and the far slice all moves outward,
    # so nothing left in the field.
    out = cmt.apply_rsds(
        field=field,
        los_displacement=losv,
        distance=distance,
        n_subcells=1,
    )

    np.all(out == 0)

    # Now, they both point inwards, so we should end up with all ones.
    losv = np.ones((2, 10)) * un.pixel
    losv[0] = -1 * un.pixel

    out = cmt.apply_rsds(
        field=field,
        los_displacement=losv,
        distance=distance,
        n_subcells=1,
    )

    np.all(out == 1)

    # In fact, any velocity field that is constant should keep the out constant.
    losv = 3 * np.ones((2, 10)) * un.pixel

    out = cmt.apply_rsds(
        field=field,
        los_displacement=losv,
        distance=distance,
        n_subcells=1,
    )

    np.all(out == 1)

    # Here, every pixel goes to the middle, but this *just* includes the two pixels
    # we put in.
    losv = (
        np.array(
            [
                -0.5 * np.ones(10),
                0.5 * np.ones(10),
            ]
        )
        * un.pixel
    )

    out = cmt.apply_rsds(
        field=field,
        los_displacement=losv,
        distance=distance,
        n_subcells=1,
    )

    np.all(out == 1)

    out = cmt.apply_rsds(
        field=field,
        los_displacement=losv,
        distance=distance,
        n_subcells=4,
    )

    np.all(out == 1)

    # Now, change the distance ever so slightly, so we use the other interpolator.
    field = np.ones((7, 10))

    losv = np.zeros((7, 10)) * un.pixel
    distance = np.array([10, 11.001, 12, 13, 14, 15, 16]) * un.pixel

    out = cmt.apply_rsds(
        field=field,
        los_displacement=losv,
        distance=distance,
        n_subcells=1,
    )

    np.allclose(out, 1, atol=2e-3)


def test_vector_field_nonzero_origin() -> None:
    """Test that moving the origin doesn't change the LoS component of a static vfield.

    This was originally a bug -- while the field is homogeneous, and so the origin
    can be arbitrary, when taking the line-of-sight component of a field, we have to make
    sure that the coordinates are correct with respect to an actual origin, otherwise
    the dot product of the vector field with the coordinates is wrong.
    """
    vx = np.ones((10, 10, 10))
    vy = np.zeros((10, 10, 10))
    vz = np.zeros((10, 10, 10))

    latitude = np.zeros(21)
    longitude = np.linspace(0, 2 * np.pi, 21)
    interpolator = cmt.make_lightcone_slice_interpolator(
        longitude=longitude,
        latitude=latitude,
        distance_to_shell=5.5,
    )

    interpolator_origin = cmt.make_lightcone_slice_interpolator(
        longitude=longitude,
        latitude=latitude,
        distance_to_shell=5.5,
        origin=(1, 2, 3),
    )

    los = next(cmt.make_lightcone_slice_vector_field([[vx, vy, vz]], interpolator))

    los_origin = next(
        cmt.make_lightcone_slice_vector_field([[vx, vy, vz]], interpolator_origin)
    )

    np.testing.assert_allclose(los, los_origin, atol=1e-6, rtol=1e-6)


def test_transform_with_different_origin_types() -> None:
    """Test that passing the same origin using different types works as expected."""
    latitude = np.zeros(21)
    longitude = np.linspace(0, 2 * np.pi, 21)

    tuple_origin = cmt.transform_to_pixel_coords(
        comoving_radius=10,
        latitude=latitude,
        longitude=longitude,
        origin=(1, 2, 3),
    )

    array_origin = cmt.transform_to_pixel_coords(
        comoving_radius=10,
        latitude=latitude,
        longitude=longitude,
        origin=np.array([1, 2, 3]),
    )

    np.testing.assert_equal(tuple_origin, array_origin)



"""Cosmotile."""

from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Generator
from typing import Literal
from typing import Sequence

import numpy as np
from astropy import units as un
from astropy.cosmology import FLRW
from astropy.cosmology import Planck18
from astropy_healpix import HEALPix
from scipy.interpolate import RectBivariateSpline
from scipy.interpolate import RegularGridInterpolator
from scipy.ndimage import map_coordinates
from scipy.spatial.transform import Rotation

from . import _version
from .cic import cloud_in_cell_los


__version__ = _version.version

_LENGTH = "length"


def get_distance_to_shell_from_redshift(
    z: float, cell_size: un.Quantity[_LENGTH], cosmo: FLRW = Planck18
) -> un.Quantity[un.pixel]:
    """Get a distance to a shell, in units of cell size, from a given redshift.

    Parameters
    ----------
    z
        The redshift
    cell_size
        The resolution of the coeval simulation, in comoving units.
    cosmo
        The astropy cosmology.

    Returns
    -------
    distance
        The distance, in units of pixels, to the shell.
    """
    return (cosmo.comoving_distance(z)).to(
        un.pixel, un.pixel_scale(cell_size / un.pixel)
    )


def make_lightcone_slice_interpolator(
    *,
    latitude: np.ndarray,
    longitude: np.ndarray,
    distance_to_shell: float,
    interpolation_order: int = 1,
    origin: np.ndarray | tuple[float, float, float] | None = None,
    rotation: Rotation | None = None,
) -> partial[np.ndarray]:
    """
    Create a callable interpolator for a lightcone slice.

    Parameters
    ----------
    latitude
        An array of latitude coordinates onto which to tile the box. In radians from
        -pi/2 to pi/2
    longitude
        An array, same size as latitude, of longitude coordinates onto which to tile the
        box. In radians from 0 to 2pi.
    distance_to_shell
        The distance to the spherical shell onto which to interpolate, in units of
        the cell-size of the coeval box(es) you wish to interpolate.
    interpolation_order
        The order of interpolation. Must be in the range 0-5.
    origin
        Define the location of the centre of the spherical shell, assuming that the
        (0,0,0) pixel of the coeval box is at (0,0,0) in cartesian coordinates.
    rotation
        The rotation by which to rotate the spherical coordinates before interpolation.
        This is done before shifting the origin, and is equivalent to rotating the
        coeval box beforing tiling it.

    Returns
    -------
    interpolator
        A callable that takes a 3D array of coeval values and returns a 2D array of
        interpolated values on a redshift slice.
    """
    if distance_to_shell <= 0:
        raise ValueError("distance_to_shell must be positive")

    if interpolation_order < 0 or interpolation_order > 5:
        raise ValueError("interpolation_order must be in the range 0-5")

    if not isinstance(interpolation_order, int):
        raise TypeError("interpolation_order must be an integer")

    if isinstance(origin, (tuple, list)):
        origin = np.array(origin)

    if origin is not None and origin.shape != (3,):
        raise ValueError("origin must be a sequence of length 3")

    pixel_coords = transform_to_pixel_coords(
        comoving_radius=distance_to_shell,
        latitude=latitude,
        longitude=longitude,
        origin=origin,
        rotation=rotation,
    )

    coordmap = partial(
        map_coordinates,
        coordinates=pixel_coords,
        order=interpolation_order,
        mode="grid-wrap",  # this wraps each dimension.
        prefilter=False,
    )

    # Save the origin to the coordmap because it's useful for getting
    # line-of-sight vectors.
    coordmap.origin = origin

    coordmap.__name__ = "lightcone_slice_interpolator"
    coordmap.__doc__ = """Interpolate a coeval box to a lightcone slice at a given redshift.

This function is a wrapper around :func:`scipy.ndimage.map_coordinates` created by
functools.partial.

Parameters
----------
coeval
    A 3D array of float coeval values to be interpolated to the lightcone slice.

Returns
-------
lightcone_slice
    A 2D array of float interpolated values on the lightcone slice.
"""
    return coordmap


def make_lightcone_slice(
    *, coevals: Sequence[np.ndarray] | np.ndarray, **kwargs: Any
) -> Generator:
    """
    Create a lightcone slice in angular coordinates from two coeval simulations.

    Interpolates the input coeval box to angular coordinates.

    Parameters
    ----------
    coevals
        An iterable of rectangular coeval simulations to interpolate to the angular
        coordinates. Must have three dimensions (not necessarily the same size). Each
        box must have the same shape, and all are assumed to be at the same coordinates.
        Each coeval box can be a different simulated field.

    Other Parameters
    ----------------
    All other parameters are passed to :func:`make_lightcone_slice_interpolator`.

    Yields
    ------
    field
        Each interpolated field on the angular coordinates.
    """
    if isinstance(coevals, np.ndarray) and coevals.ndim == 3:
        coevals = [coevals]

    if any(cv.ndim != 3 for cv in coevals):
        raise ValueError("all coevals must have three dimensions")

    if any(cv.shape != coevals[0].shape for cv in coevals):
        raise ValueError("all coevals must have the same shape")

    coordmap = make_lightcone_slice_interpolator(**kwargs)

    for cv in coevals:
        yield coordmap(cv)


def make_lightcone_slice_vector_field(
    coeval_vector_fields: Sequence[Sequence[np.ndarray]],
    interpolator: Callable[[np.ndarray], np.ndarray],
) -> Generator:
    """
    Interpolate a 3D vector field to a lightcone slice as a line-of-sight component.

    This takes a sequence of 3D vector fields, eg. the velocity field, and interpolates
    each component to the lightcone slice. It then computes the line-of-sight component
    of each interpolated vector field, where positive values are oriented towards the
    observer.

    Parameters
    ----------
    coeval_vector_fields
        An iterable of 3D vector fields to interpolate to the lightcone slice. Each
        vector field must be an iterable of 3 3D arrays, each of the same shape.
    interpolator
        A callable that takes a 3D array of coeval values and returns a 2D array of
        interpolated values on a redshift slice. This should be created by
        :func:`make_lightcone_slice_interpolator` using the properties of the
        coeval vector fields.

    Yields
    ------
    los_component
        The line-of-sight component of each interpolated vector field.
    """
    pixel_coords = interpolator.keywords["coordinates"]
    if interpolator.origin is not None:
        pixel_coords = pixel_coords - interpolator.origin[:, None]

    coord_norm = np.sqrt(np.sum(np.square(pixel_coords), axis=0))
    for i, cvf in enumerate(coeval_vector_fields):
        if len(cvf) != 3:
            raise ValueError(
                "coeval_vector_fields must be a sequence of 3-tuples. Got length "
                f"{len(cvf)} at index {i}"
            )
        if any(c.shape != cvf[0].shape for c in cvf):
            raise ValueError(
                f"all coeval vector fields must have the same shape. For index {i}, "
                f"got shapes {[c.shape for c in cvf]}"
            )
        unit = getattr(cvf[0], "unit", 1)
        cvf_interp = np.array([interpolator(c) for c in cvf]) * unit

        # Now take the dot product of the vector field with (negative) pixel coordinates to get
        # the LoS comp.
        cvf_interp *= -pixel_coords
        yield np.sum(cvf_interp, axis=0) / coord_norm


def transform_to_pixel_coords(
    *,
    comoving_radius: un.Quantity[un.pixel],
    latitude: np.ndarray,
    longitude: np.ndarray,
    origin: un.Quantity[un.pixel, (3,), float]
    | tuple[float, float, float]
    | None = None,
    rotation: Rotation | None = None,
) -> np.ndarray:
    """Transform input spherical coordinates to pixel coordinates wrt a coeval box.

    Parameters
    ----------
    comoving_radius
        The radius of the spherical coordinates (in units of the cell size).
    latitude
        An array of latitude coordinates onto which to tile the box. In radians from
        -pi/2 to pi/2
    longitude
        An array, same size as latitude, of longitude coordinates onto which to tile the
        box. In radians from 0 to 2pi.
    origin
        Define the location of the centre of the spherical shell, assuming that the
        (0,0,0) pixel of the coeval box is at (0,0,0) in cartesian coordinates.
        In units of the cell size.
    rotation
        The rotation by which to rotate the spherical coordinates before interpolation.
        This is done before shifting the origin, and is equivalent to rotating the
        coeval box beforing tiling it.
    """
    if latitude.shape != longitude.shape:
        raise ValueError("latitude and longitude must have the same shape")

    if latitude.ndim != 1:
        raise ValueError("latitude and longitude must be 1D arrays")

    if np.any(np.abs(latitude) > np.pi / 2):
        raise ValueError("latitude must be between -pi/2 and pi/2")

    if np.any((longitude < 0) | (longitude > 2 * np.pi)):
        raise ValueError("longitude must be between 0 and 2pi")

    # Get the cartesian coordinates (x, y, z) of the angular lightcone coords.
    phi = np.pi / 2 - latitude
    sinphi = np.sin(phi)
    cart_coords = comoving_radius * np.array(
        [
            sinphi * np.cos(longitude),
            sinphi * np.sin(longitude),
            np.cos(phi),
        ]
    )

    # Get a rotation matrix
    if rotation is not None:
        cart_coords = np.dot(rotation.as_matrix(), cart_coords)

    # Apply an offset transformation if desired.
    if origin is not None:
        if not isinstance(origin, np.ndarray):
            origin = np.array(origin)
        cart_coords += origin[:, None]

    return cart_coords


def make_healpix_lightcone_slice(
    nside: int, order: Literal["ring", "nested"] = "ring", **kwargs: Any
) -> Generator:
    """
    Create a healpix lightcone slice in angular coordinates.

    This is a simple wrapper around :func:`make_lightcone_slice` that sets up angular
    co-ordinates from a healpix grid.

    Parameters
    ----------
    nside
        The Nside parameter of the healpix map.
    order
        The ordering of the pixels in the healpix map.

    Other Parameters
    ----------------
    All other parameters are passed through to :func:`make_lightcone_slice`.
    """
    hp = HEALPix(nside=nside, order=order)
    lon, lat = hp.healpix_to_lonlat(np.arange(hp.npix))

    yield from make_lightcone_slice(
        latitude=lat.to_value("radian"), longitude=lon.to_value("radian"), **kwargs
    )


def apply_rsds(
    field: np.ndarray,
    los_displacement: np.ndarray,
    distance: np.ndarray,
    n_subcells: int = 4,
) -> np.ndarray:
    """Apply redshift-space distortions to a field.

    Notes
    -----
    To ensure that we cover all the slices in the field after the velocities have
    been applied, we extrapolate the densities and velocities on either end by the
    maximum velocity offset in the field.
    Then, to ensure we don't pick up cells with zero particles (after displacement),
    we interpolate the slices onto a finer regular grid (in comoving distance) and
    then displace the field on that grid, before interpolating back onto the original
    slices.

    Parameters
    ----------
    field
        The field to apply redshift-space distortions to, shape (nslices, ncoords).
    los_displacement
        The line-of-sight "apparent" displacement of the field, in pixel coordinates.
        Equal to ``v / H(z) / cell_size``.
        Positive values are towards the observer, shape ``(nslices, ncoords)``.
    distance
        The comoving distance to each slice in the field, in units of the cell size.
        shape (nslices,).
    """
    if field.shape != los_displacement.shape:
        raise ValueError("field and los_displacement must have the same shape")
    if field.shape[0] < 2:
        raise ValueError("field must have at least 2 slices")
    if field.shape[0] != distance.size:
        raise ValueError("field and distance must have the same number of slices")

    is_regular = np.allclose(np.diff(np.diff(distance)), 0.0)
    interpolator = RegularGridInterpolator if is_regular else RectBivariateSpline

    # TODO: convert these to distances...
    vmax_towards_observer = max(np.max(los_displacement[0]), 0)
    vmax_away_from_observer = min(0, np.min(los_displacement[-1]))

    smallest_slice = np.min(np.diff(distance))
    rsd_dx = smallest_slice / n_subcells

    fine_grid = np.arange(
        (distance.min() - vmax_towards_observer).to_value(rsd_dx.unit),
        (distance.max() - vmax_away_from_observer).to_value(rsd_dx.unit),
        rsd_dx.value,
    )
    if fine_grid.max() < distance.max().to_value(rsd_dx.unit):
        fine_grid = np.append(fine_grid, distance.max().to_value(rsd_dx.unit))

    ang_coords = np.arange(field.shape[1])
    if is_regular:
        X, Y = np.meshgrid(fine_grid, ang_coords, indexing="ij")
        grid = (X.flatten(), Y.flatten())
        fine_field = interpolator(
            (distance, ang_coords), field, bounds_error=False, fill_value=None
        )(grid).reshape(X.shape)
        fine_rsd = interpolator(
            (distance, ang_coords),
            los_displacement / rsd_dx,
            bounds_error=False,
            fill_value=None,
        )(grid).reshape(X.shape)
    else:
        fine_field = interpolator(distance, ang_coords, field)(fine_grid, ang_coords)
        fine_rsd = interpolator(distance, ang_coords, los_displacement / rsd_dx)(
            fine_grid, ang_coords
        )
    fine_field = cloud_in_cell_los(fine_field, fine_rsd)

    X, Y = np.meshgrid(distance, ang_coords, indexing="ij")

    return RegularGridInterpolator((fine_grid, ang_coords), fine_field)(
        (X.flatten(), Y.flatten())
    ).reshape(X.shape)



"""Tests of the cloud-in-cell function."""
import numpy as np
import pytest

from cosmotile.cic import cloud_in_cell_coeval as cloud_in_cell
from cosmotile.cic import cloud_in_cell_los as cic_los


def test_cic_bad_input() -> None:
    """Test that different-shaped inputs to CIC raise appropriate error."""
    fld = np.zeros((10, 10, 10))
    dx = np.zeros_like(fld)
    dy = np.zeros_like(fld)
    dz = np.zeros_like(fld)

    with pytest.raises(
        ValueError, match="Field and displacement must have the same shape."
    ):
        cloud_in_cell(fld, dx, dy, dz[1:])


def test_cic_dx() -> None:
    """Test the cloud-in-cell function."""
    x = np.zeros((10, 10, 10))
    x[::2] = 1

    dx = np.zeros_like(x)
    dx[::2] = 1.0
    dy = np.zeros_like(x)
    dz = np.zeros_like(x)

    out = cloud_in_cell(x, dx, dy, dz)

    assert np.allclose(out[1::2], 1.0)

    out = cloud_in_cell(x, -dx, dy, dz)
    assert np.allclose(out[1::2], 1.0)

    out = cloud_in_cell(x, -1.5 * dx, dy, dz)
    assert np.allclose(out, 0.5)


def test_cic_dy() -> None:
    """Test the cloud-in-cell function."""
    x = np.zeros((10, 10, 10))
    x[:, ::2] = 1

    dx = np.zeros_like(x)
    dy = np.zeros_like(x)
    dy[:, ::2] = 1.0
    dz = np.zeros_like(x)

    out = cloud_in_cell(x, dx, dy, dz)
    assert np.allclose(out[:, 1::2], 1.0)
    out = cloud_in_cell(x, dx, -dy, dz)
    assert np.allclose(out[:, 1::2], 1.0)
    out = cloud_in_cell(x, dx, -1.5 * dy, dz)
    assert np.allclose(out, 0.5)


def test_cic_dz() -> None:
    """Test the cloud-in-cell function."""
    x = np.zeros((10, 10, 10))
    x[:, :, ::2] = 1

    dx = np.zeros_like(x)
    dy = np.zeros_like(x)
    dz = np.zeros_like(x)
    dz[:, :, ::2] = 1.0

    out = cloud_in_cell(x, dx, dy, dz)
    assert np.allclose(out[:, :, 1::2], 1.0)
    out = cloud_in_cell(x, dx, dy, -dz)
    assert np.allclose(out[:, :, 1::2], 1.0)
    out = cloud_in_cell(x, dx, dy, -1.5 * dz)
    assert np.allclose(out, 0.5)


def test_cic_los() -> None:
    """Test the cloud-in-cell line-of-sight function."""
    field = np.zeros((10, 20))
    los = np.zeros((11, 20))

    with pytest.raises(
        ValueError, match="Field and displacement must have the same shape."
    ):
        cic_los(field, los)



"""Nox sessions."""
import os
import shutil
import sys
from pathlib import Path

import nox


package = "cosmotile"
python_versions = ["3.11", "3.10", "3.9", "3.8"]
nox.needs_version = ">= 2021.6.6"
nox.options.sessions = (
    "mypy",
    "tests",
    "xdoctest",
    "docs-build",
    "tests-nojit",
)


@nox.session(python=python_versions)
def mypy(session: nox.Session) -> None:
    """Type-check using mypy."""
    args = session.posargs or ["src", "tests", "docs/conf.py"]
    session.install(".")
    session.install("mypy", "pytest")
    session.run("mypy", *args)
    if not session.posargs:
        session.run("mypy", f"--python-executable={sys.executable}", "noxfile.py")


@nox.session(python=python_versions)
def tests(session: nox.Session) -> None:
    """Run the test suite."""
    session.install(".[all]")
    session.install("coverage[toml]", "pytest", "pygments")
    try:
        session.run("coverage", "run", "--parallel", "-m", "pytest", *session.posargs)
    finally:
        if session.interactive:
            session.notify("coverage", posargs=[])


@nox.session(python=python_versions, name="tests-nojit")
def tests_nojit(session: nox.Session) -> None:
    """Run the test suite."""
    session.install(".")
    session.install("coverage[toml]", "pytest", "pygments")
    try:
        session.run(
            "coverage", "run", "--parallel", "-m", "pytest", "-s", *session.posargs
        )
    finally:
        if session.interactive:
            session.notify("coverage", posargs=[])


@nox.session(python=python_versions[0])
def coverage(session: nox.Session) -> None:
    """Produce the coverage report."""
    args = session.posargs or ["report"]
    session.install(".")
    session.install("coverage[toml]")

    if not session.posargs and any(Path().glob(".coverage.*")):
        session.run("coverage", "combine")

    session.run("coverage", *args)


@nox.session(python=python_versions)
def xdoctest(session: nox.Session) -> None:
    """Run examples with xdoctest."""
    if session.posargs:
        args = [package, *session.posargs]
    else:
        args = [f"--modname={package}", "--command=all"]
        if "FORCE_COLOR" in os.environ:
            args.append("--colored=1")

    session.install(".")
    session.install("xdoctest[colors]")
    session.run("python", "-m", "xdoctest", *args)


@nox.session(name="docs-build", python=python_versions[0])
def docs_build(session: nox.Session) -> None:
    """Build the documentation."""
    args = session.posargs or ["docs", "docs/_build"]
    if not session.posargs and "FORCE_COLOR" in os.environ:
        args.insert(0, "--color")

    session.install(".")
    session.install("sphinx", "sphinx-click", "furo", "myst-parser")

    build_dir = Path("docs", "_build")
    if build_dir.exists():
        shutil.rmtree(build_dir)

    session.run("sphinx-build", *args)


@nox.session(python=python_versions[0])
def docs(session: nox.Session) -> None:
    """Build and serve the documentation with live reloading on file changes."""
    args = session.posargs or ["--open-browser", "docs", "docs/_build"]
    session.install(".")
    session.install("sphinx", "sphinx-autobuild", "sphinx-click", "furo", "myst-parser")

    build_dir = Path("docs", "_build")
    if build_dir.exists():
        shutil.rmtree(build_dir)

    session.run("sphinx-autobuild", *args)



"""An implementation of Cloud-in-Cell interpolation in 3D."""

from __future__ import annotations

import warnings

import numpy as np


try:
    from numba import njit

    NUMBA = True
except ImportError:
    NUMBA = False


def cloud_in_cell_coeval(
    field: np.ndarray, dx: np.ndarray, dy: np.ndarray, dz: np.ndarray
) -> np.ndarray:
    """
    Interpolates a point in 3D space.

    Note that the implementation was derived largely from
    https://astro.uchicago.edu/~andrey/Talks/PM/pm.pdf, specially slide 13.

    Parameters
    ----------
    field
        The original field before displacement.
    dx, dy, dz
        Displacement of each coordinate in the field in the x,y, and z direction.
        The displacement must be in units of the cell size. Displacements can represent,
        for example, redshift-space distortions (RSDs) or peculiar velocities. If
        velocities are in km / s, then input displacement as v / H(z) / cell_size.
    """
    if not NUMBA:  # pragma: no cover
        warnings.warn("Install numba for a speedup of cloud_in_cell", stacklevel=2)

    if not field.shape == dx.shape == dy.shape == dz.shape:
        raise ValueError("Field and displacement must have the same shape.")

    out = np.zeros_like(field)

    nx, ny, nz = dx.shape
    for ii in range(nx):
        for jj in range(ny):
            for kk in range(nz):
                weight = field[ii, jj, kk]

                # Get the offset of this grid cell
                ddx = dx[ii, jj, kk]
                ddy = dy[ii, jj, kk]
                ddz = dz[ii, jj, kk]

                # adding a value of nx pre-mod ensures we are still within the range [0, nx]
                x = (ii + ddx + nx) % nx
                y = (jj + ddy + ny) % ny
                z = (kk + ddz + nz) % nz

                i = int(x)
                j = int(y)
                k = int(z)

                ip = (i + 1) % nx
                jp = (j + 1) % ny
                kp = (k + 1) % nz

                tx = (i + 1) - x
                ty = (j + 1) - y
                tz = (k + 1) - z

                ddx = 1 - tx
                ddy = 1 - ty
                ddz = 1 - tz

                # Do the trilinear interpolation to surrounding 8 cells
                out[i, j, k] += tx * ty * tz * weight
                out[ip, j, k] += ddx * ty * tz * weight
                out[i, jp, k] += tx * ddy * tz * weight
                out[ip, jp, k] += ddx * ddy * tz * weight
                out[i, j, kp] += tx * ty * ddz * weight
                out[ip, j, kp] += ddx * ty * ddz * weight
                out[i, jp, kp] += tx * ddy * ddz * weight
                out[ip, jp, kp] += ddx * ddy * ddz * weight

    return out


def cloud_in_cell_los(
    field: np.ndarray,
    delta_los: np.ndarray,
) -> np.ndarray:
    """
    Interpolate in the line-of-sight direction using cloud-in-cell algorithm.

    Note that the implementation was derived largely from
    https://astro.uchicago.edu/~andrey/Talks/PM/pm.pdf, specially slide 13.

    Notes
    -----
    The ``field`` here is assumed to be regularly spaced in comoving distance along the
    line-of-sight (the angular coordinates can be arbitrarily arranged). For each angular
    point, we first displace the regular grid along the line-of-sight by ``delta_los``,
    to create a a new, non-regular grid (which we can consider to be "particles"). We
    then use the regular cloud-in-cell interpolation to interpolate the particles back
    on to the regular grid.

    Parameters
    ----------
    field
        The regularly-spaced (along LoS) field before displacement by delta_los.
        Shape ``(nlos_slices, nangles)``.
    delta_los
        Displacement of each coordinate in the field along the LoS.
        The displacement must be in units of the regular grid size, i.e.
        ``v / H(z) / grid_resolution``. Same shape as ``field``.
    """
    if not NUMBA:  # pragma: no cover
        warnings.warn("Install numba for a speedup of cloud_in_cell", stacklevel=2)

    if field.shape != delta_los.shape:
        raise ValueError("Field and displacement must have the same shape.")

    out = np.zeros_like(field)

    nslice, nangles = delta_los.shape
    for ii in range(nslice):
        weight = field[ii]

        # Get the offset of this grid cell
        ddx = delta_los[ii]
        x = ii + ddx

        i = x.astype(np.int32)
        ip = i + 1

        tx = ip - x
        ddx = 1 - tx

        for jj in range(nangles):
            if 0 <= i[jj] < nslice:
                out[i[jj], jj] += tx[jj] * weight[jj]
            if 0 <= ip[jj] < nslice:
                out[ip[jj], jj] += ddx[jj] * weight[jj]
    return out


if NUMBA:
    cloud_in_cell_coeval_nojit = cloud_in_cell_coeval
    cloud_in_cell_coeval = njit(cloud_in_cell_coeval)
    cloud_in_cell_los_nojit = cloud_in_cell_los
    cloud_in_cell_los = njit(cloud_in_cell_los)
